import logging
from pathlib import Path

# TODO(pankaj) In desparate need of refactoring into separate library
try:
    from helpers.types import Action, ModelCodePatch
except ModuleNotFoundError:
    from truss.templates.control.control.helpers.types import Action, ModelCodePatch


def apply_model_code_patch(
    model_code_dir: Path,
    patch: ModelCodePatch,
    logger: logging.Logger,
):
    logger.debug(f"Applying model code patch {patch.to_dict()}")
    action = patch.action
    filepath: Path = model_code_dir / patch.path
    if action in [Action.ADD, Action.UPDATE]:
        filepath.parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"Updating file {filepath}")
        with filepath.open("w") as file:
            content = patch.content
            if content is None:
                raise ValueError(
                    "Invalid patch: content of a model code update patch should not be None."
                )
            file.write(content)

    elif action == Action.REMOVE:
        if not filepath.exists():
            logger.warning(f"Could not delete file {filepath}: not found.")
        else:
            logger.info(f"Deleting file {filepath}")
            filepath.unlink()
    else:
        raise ValueError(f"Unknown model code patch action {action}")



from typing import Any, Dict

from fastapi import APIRouter
from fastapi.responses import JSONResponse
from helpers.errors import ModelLoadFailed, ModelNotReady
from httpx import URL, ConnectError
from starlette.requests import Request
from starlette.responses import Response
from tenacity import Retrying, retry_if_exception_type, stop_after_attempt, wait_fixed

INFERENCE_SERVER_START_WAIT_SECS = 60


control_app = APIRouter()


@control_app.get("/")
def index():
    return {}


async def proxy(request: Request):
    inference_server_process_controller = (
        request.app.state.inference_server_process_controller
    )
    client = request.app.state.proxy_client
    url = URL(path=request.url.path, query=request.url.query.encode("utf-8"))
    rp_req = client.build_request(
        request.method, url, headers=request.headers.raw, content=await request.body()
    )

    # Wait a bit for inference server to start
    for attempt in Retrying(
        retry=(
            retry_if_exception_type(ConnectError)
            | retry_if_exception_type(ModelNotReady)
        ),
        stop=stop_after_attempt(INFERENCE_SERVER_START_WAIT_SECS),
        wait=wait_fixed(1),
    ):
        with attempt:
            try:
                if (
                    inference_server_process_controller.is_inference_server_intentionally_stopped()
                ):
                    raise ModelLoadFailed("Model load failed")
                resp = await client.send(rp_req)

                if _is_model_not_ready(resp):
                    raise ModelNotReady("Model has started running, but not ready yet.")
            except ConnectionError as exp:
                # This check is a bit expensive so we don't do it before every request, we
                # do it only if request fails with connection error. If the inference server
                # process is running then we continue waiting for it to start (by retrying),
                # otherwise we bail.
                if (
                    inference_server_process_controller.inference_server_ever_started()
                    and not inference_server_process_controller.is_inference_server_running()
                ):
                    error_msg = "It appears your model has stopped running. This often means' \
                        ' it crashed and may need a fix to get it running again."
                    return JSONResponse(error_msg, 503)
                raise exp

    response = Response(resp.content, resp.status_code, resp.headers)
    return response


control_app.add_route("/v1/{path:path}", proxy, ["GET", "POST"])


@control_app.post("/control/patch")
async def patch(request: Request) -> Dict[str, str]:
    request.app.state.logger.info("Patch request received.")
    patch_request = await request.json()
    request.app.state.inference_server_controller.apply_patch(patch_request)
    request.app.state.logger.info("Patch applied successfully")
    return {"msg": "Patch applied successfully"}


@control_app.get("/control/truss_hash")
async def truss_hash(request: Request) -> Dict[str, Any]:
    t_hash = request.app.state.inference_server_controller.truss_hash()
    return {"result": t_hash}


@control_app.post("/control/restart_inference_server")
async def restart_inference_server(request: Request) -> Dict[str, str]:
    request.app.state.inference_server_controller.restart()

    return {"msg": "Inference server started successfully"}


@control_app.get("/control/has_partially_applied_patch")
async def has_partially_applied_patch(request: Request) -> Dict[str, Any]:
    app_has_partially_applied_patch = (
        request.app.state.inference_server_controller.has_partially_applied_patch()
    )
    return {"result": app_has_partially_applied_patch}


@control_app.post("/control/stop_inference_server")
async def stop_inference_server(request: Request) -> Dict[str, str]:
    request.app.state.inference_server_controller.stop()
    return {"msg": "Inference server stopped successfully"}


def _is_model_not_ready(resp) -> bool:
    return (
        resp.status_code == 503
        and resp.content is not None
        and "model is not ready" in resp.content.decode("utf-8")
    )



#!/usr/bin/env python3
import logging
import subprocess
import sys
import venv
from pathlib import Path
from typing import Any, List, Optional

from truss.contexts.image_builder.serving_image_builder import ServingImageBuilder
from truss.contexts.local_loader.docker_build_emulator import DockerBuildEmulator
from truss.contexts.local_loader.truss_file_syncer import TrussFilesSyncer
from truss.patch.local_truss_patch_applier import LocalTrussPatchApplier
from truss.util.path import build_truss_shadow_target_directory
from yaspin import yaspin


class VenvBuilder(venv.EnvBuilder):
    """Virtual Environment Builder

    This class handles setting up a virtual environment in `venv_dir`.
    """

    def __init__(self, venv_dir: Path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.context: Any = None
        self.venv_dir = venv_dir

    def post_setup(self, context):
        """Implement base class `post_setup` hook to store venv context."""
        self.context = context

    def create_with_requirements(self, req_files: List[str]) -> None:
        """Create virtualenv with the requirements specified in the files provided.

        Args:
            `req_file` (List[str]): List of requirement files as subpaths to `self.venv_dir`.
        """
        with yaspin(text="Creating virtual environment") as spinner:
            env_dir = str(self.venv_dir / ".env")
            super().create(env_dir)
            spinner.ok("✅")
        with yaspin(text="Installing depedencies") as spinner:
            self._upgrade_pip()
            for req_file in req_files:
                self._install_pip_requirements(self.venv_dir / req_file)
            spinner.ok("✅")

    def _upgrade_pip(self):
        """Helper function to upgrade pip."""
        subprocess.check_call(
            [
                self.context.env_exe,
                "-m",
                "pip",
                "install",
                "--upgrade",
                "pip",
            ],
            stdout=subprocess.PIPE,
        )

    def _install_pip_requirements(self, req_file: Path):
        """Helper function to install requirements from file."""
        if req_file.exists():
            pip_install_command = [
                self.context.env_exe,
                "-m",
                "pip",
                "install",
                "-r",
                str(req_file.absolute()),
            ]
            subprocess.check_call(
                pip_install_command,
                stdout=subprocess.PIPE,
            )


class LocalServerLoader:
    """Handle the setup and loading of truss server locally."""

    def __init__(
        self,
        truss_path: Path,
        context_builder: ServingImageBuilder,
        port: Optional[int] = None,
    ) -> None:
        self.truss_path = truss_path
        self.context_builder = context_builder
        self.port = port or 8080

    def watch(
        self,
        build_dir: Optional[Path] = None,
        venv_dir: Optional[Path] = None,
    ):
        """Run the server and watch for changes"""

        def _prep_or_create_dir(location: Optional[Path], stub: str) -> Path:
            if location is None:
                location = build_truss_shadow_target_directory(stub, self.truss_path)
            else:
                if not location.exists():
                    location.mkdir(parents=True)
            return location

        build_dir = _prep_or_create_dir(build_dir, "build")
        venv_dir = _prep_or_create_dir(venv_dir, "server_venv")

        with yaspin(text="Preparing truss context") as spinner:
            self.context_builder.prepare_image_build_dir(build_dir)
            dockerfile_path = build_dir / "Dockerfile"
            docker_build_emulator = DockerBuildEmulator(dockerfile_path, build_dir)
            build_result = docker_build_emulator.run(venv_dir)
            spinner.ok("✅")

        venv_builder = VenvBuilder(venv_dir, with_pip=True)
        requirements_files = [
            "app/requirements.txt",
            "requirements.txt",
        ]
        venv_builder.create_with_requirements(requirements_files)

        TrussFilesSyncer(
            self.truss_path,
            LocalTrussPatchApplier(
                venv_dir / "app/",
                venv_builder.context.env_exe,
                logging.Logger(__name__),
            ),
        ).start()

        subprocess.check_call(
            [
                venv_builder.context.env_exe,
                "-m",
                "uvicorn",
                "local_inference_server:app",
                "--reload",
                "--port",
                str(self.port),
            ],
            cwd=str(venv_dir / str(build_result.workdir).replace("/", "", 1)),
            env={
                **build_result.env,
                "SETUP_JSON_LOGGER": "False",
            },
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )



import asyncio
import concurrent.futures
import logging
import multiprocessing
import socket
from typing import List, Optional

import shared.util as utils
import uvicorn
from fastapi import FastAPI

FORMAT = "%(asctime)s.%(msecs)03d %(name)s %(levelname)s [%(funcName)s():%(lineno)s] %(message)s"
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
logging.basicConfig(level=logging.INFO, format=FORMAT, datefmt=DATE_FORMAT)


class UvicornCustomServer(multiprocessing.Process):
    def __init__(
        self, config: uvicorn.Config, sockets: Optional[List[socket.socket]] = None
    ):
        super().__init__()
        self.sockets = sockets
        self.config = config

    def stop(self):
        self.terminate()

    def run(self):
        server = uvicorn.Server(config=self.config)
        asyncio.run(server.serve(sockets=self.sockets))


def start_uvicorn_server(application: FastAPI, host: str = "*", port: int = 8080):
    cfg = uvicorn.Config(
        application,
        host=host,
        port=port,
        workers=1,
    )

    max_asyncio_workers = min(32, utils.cpu_count() + 4)
    logging.info(f"Setting max asyncio worker threads as {max_asyncio_workers}")
    # Call this so uvloop gets used
    cfg.setup_event_loop()
    asyncio.get_event_loop().set_default_executor(
        concurrent.futures.ThreadPoolExecutor(max_workers=max_asyncio_workers)
    )

    async def serve():
        serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        serversocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        serversocket.bind((cfg.host, cfg.port))
        serversocket.listen(5)

        logging.info(f"starting uvicorn with {cfg.workers} workers")
        for _ in range(cfg.workers):
            server = UvicornCustomServer(config=cfg, sockets=[serversocket])
            server.start()

    async def servers_task():
        servers = [serve()]
        await asyncio.gather(*servers)

    asyncio.run(servers_task())



import logging
from pathlib import Path
from threading import Thread
from typing import List, Optional

from truss.constants import CONFIG_FILE
from truss.patch.calc_patch import calc_config_patches
from truss.patch.local_truss_patch_applier import LocalTrussPatchApplier
from truss.templates.control.control.helpers.types import (
    Action,
    ModelCodePatch,
    Patch,
    PatchType,
)
from truss.truss_config import TrussConfig
from truss.truss_spec import TrussSpec

# from truss.util.path import is_ignored, load_trussignore_patterns
from watchfiles import Change, watch

OP_TO_ACTION = {
    Change.added: Action.ADD,
    Change.deleted: Action.REMOVE,
    Change.modified: Action.UPDATE,
}


class TrussPatchEmitter:
    """Callable that emits an optional list of patches to apply to the running local given a file change"""

    def __init__(
        self,
        truss_dir: Path,
        logger: logging.Logger,
    ) -> None:
        self._truss_dir = truss_dir
        self._config = TrussConfig.from_yaml(self._truss_dir / CONFIG_FILE)
        self._logger = logger

    def __call__(self, op: Change, path: Path) -> Optional[List[Patch]]:
        truss_spec = TrussSpec(self._truss_dir)
        model_module_path = str(
            truss_spec.model_module_dir.relative_to(self._truss_dir)
        )
        if str(path).startswith(model_module_path):
            rel_path = str(path.relative_to(model_module_path))
            return [
                Patch(
                    type=PatchType.MODEL_CODE,
                    body=ModelCodePatch(
                        OP_TO_ACTION[op],
                        rel_path,
                        (self._truss_dir / path).read_text(),
                    ),
                )
            ]
        if str(path) == CONFIG_FILE:
            new_config = TrussConfig.from_yaml(self._truss_dir / CONFIG_FILE)
            config_patches = calc_config_patches(self._config, new_config)
            self._config = new_config
            return config_patches or None
        return None


class TrussFilesSyncer(Thread):
    """Daemon thread that watches for changes in the user's Truss and syncs to running service."""

    def __init__(self, watch_path: Path, patch_applier: LocalTrussPatchApplier) -> None:
        super().__init__(daemon=True)
        self._logger = logging.Logger(__name__)
        self.watch_path = watch_path
        self.emit_patches = TrussPatchEmitter(self.watch_path, self._logger)
        self.patch_applier = patch_applier
        # self.watch_filter = lambda _, path: not is_ignored(
        #     Path(path),
        #     load_trussignore_patterns(),
        # )

    def run(self) -> None:
        """Watch for files in background and apply appropriate patches."""
        for changes in watch(self.watch_path):  # watch_filter=self.watch_filter):
            for change in changes:
                op, path = change
                rel_path = Path(path).relative_to(self.watch_path.resolve())
                patches = self.emit_patches(op, rel_path)
                if patches:
                    self.patch_applier(patches)



import os

import yaml
from common.truss_server import TrussServer  # noqa: E402

CONFIG_FILE = "config.yaml"

env_port = int(os.environ.get("INFERENCE_SERVER_PORT", "8080"))
setup_json_logger = bool(os.environ.get("SETUP_JSON_LOGGER", "True") == "True")
with open(CONFIG_FILE, encoding="utf-8") as config_file:
    config = yaml.safe_load(config_file)

server = TrussServer(
    http_port=env_port,
    config=config,
    setup_json_logger=setup_json_logger,
)
app = server.create_application()



import subprocess
from pathlib import Path

import pytest
from truss.constants import SUPPORTED_PYTHON_VERSIONS
from truss.tests.conftest import CUSTOM_MODEL_USING_EXTERNAL_PACKAGE_CODE
from truss.tests.test_testing_utilities_for_other_tests import ensure_kill_all
from truss.tests.test_truss_handle import (
    verify_environment_variable_on_container,
    verify_python_requirement_installed_on_container,
    verify_python_requirement_not_installed_on_container,
    verify_system_package_installed_on_container,
    verify_system_requirement_not_installed_on_container,
)
from truss.truss_handle import TrussHandle


def current_num_docker_images(th: TrussHandle) -> int:
    return len(th.get_all_docker_images())


@pytest.fixture
def control_model_handle_tag_tuple(
    custom_model_control,
) -> tuple[Path, TrussHandle, str]:
    th = TrussHandle(custom_model_control)
    tag = "test-docker-custom-model-control-tag:0.0.1"
    return (custom_model_control, th, tag)


@pytest.mark.integration
@pytest.mark.parametrize(
    "binary, python_version",
    [
        (binary, python_version)
        for binary in [True, False]
        for python_version in SUPPORTED_PYTHON_VERSIONS
    ],
)
def test_control_truss_model_code_patch(
    binary, python_version, control_model_handle_tag_tuple
):
    custom_model_control, th, tag = control_model_handle_tag_tuple
    th.update_python_version(python_version)

    def predict_with_updated_model_code():
        new_model_code = """
class Model:
    def predict(self, model_input):
        return [2 for i in model_input]
"""
        model_code_file_path = custom_model_control / "model" / "model.py"
        with model_code_file_path.open("w") as model_code_file:
            model_code_file.write(new_model_code)
        return th.docker_predict([1], tag=tag, binary=binary)

    with ensure_kill_all():
        result = th.docker_predict([1], tag=tag, binary=binary)
        assert result[0] == 1
        orig_num_truss_images = len(th.get_all_docker_images())

        result = predict_with_updated_model_code()
        assert result[0] == 2
        assert orig_num_truss_images == current_num_docker_images(th)


@pytest.mark.integration
@pytest.mark.parametrize(
    "binary, python_version",
    [
        (binary, python_version)
        for binary in [True, False]
        for python_version in SUPPORTED_PYTHON_VERSIONS
    ],
)
def test_control_truss_empty_dir_patch(
    binary, python_version, control_model_handle_tag_tuple
):
    custom_model_control, th, tag = control_model_handle_tag_tuple
    th.update_python_version(python_version)

    def predict_with_added_empty_directory():
        # Adding empty directory should work
        (custom_model_control / "model" / "dir").mkdir()
        return th.docker_predict([1], tag=tag, binary=binary)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag, binary=binary)
        orig_num_truss_images = len(th.get_all_docker_images())

        predict_with_added_empty_directory()
        assert orig_num_truss_images == current_num_docker_images(th)


@pytest.mark.integration
@pytest.mark.parametrize(
    "binary, python_version",
    [
        (binary, python_version)
        for binary in [True, False]
        for python_version in SUPPORTED_PYTHON_VERSIONS
    ],
)
def test_control_truss_unpatchable(
    binary, python_version, control_model_handle_tag_tuple
):
    custom_model_control, th, tag = control_model_handle_tag_tuple
    th.update_python_version(python_version)

    def predict_with_unpatchable_change():
        # Changes that are not expressible with patch should also work
        # Changes to data dir are not currently patch expressible
        (custom_model_control / "data" / "dummy").touch()
        return th.docker_predict([1], tag=tag, binary=binary)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag, binary=binary)
        orig_num_truss_images = len(th.get_all_docker_images())

        predict_with_unpatchable_change()
        assert current_num_docker_images(th) == orig_num_truss_images + 1


@pytest.mark.integration
@pytest.mark.parametrize(
    "binary, python_version",
    [
        (binary, python_version)
        for binary in [True, False]
        for python_version in SUPPORTED_PYTHON_VERSIONS
    ],
)
def test_control_truss_python_sys_req_patch(
    binary, python_version, control_model_handle_tag_tuple
):
    _, th, tag = control_model_handle_tag_tuple
    th.update_python_version(python_version)

    def predict_with_python_requirement_added(req: str):
        th.add_python_requirement(req)
        return th.docker_predict([1], tag=tag, binary=binary)

    def predict_with_python_requirement_removed(req):
        th.remove_python_requirement(req)
        return th.docker_predict([1], tag=tag, binary=binary)

    def predict_with_system_requirement_added(pkg):
        th.add_system_package(pkg)
        return th.docker_predict([1], tag=tag, binary=binary)

    def predict_with_system_requirement_removed(pkg):
        th.remove_system_package(pkg)
        return th.docker_predict([1], tag=tag, binary=binary)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag, binary=binary)
        orig_num_truss_images = len(th.get_all_docker_images())

        container = th.get_running_serving_container_ignore_hash()

        python_req = "pydot"
        predict_with_python_requirement_added(python_req)
        assert current_num_docker_images(th) == orig_num_truss_images
        verify_python_requirement_installed_on_container(container, python_req)

        predict_with_python_requirement_removed(python_req)
        assert current_num_docker_images(th) == orig_num_truss_images
        verify_python_requirement_not_installed_on_container(container, python_req)

        system_pkg = "jq"
        predict_with_system_requirement_added(system_pkg)
        assert current_num_docker_images(th) == orig_num_truss_images
        verify_system_package_installed_on_container(container, system_pkg)

        predict_with_system_requirement_removed(system_pkg)
        assert current_num_docker_images(th) == orig_num_truss_images
        verify_system_requirement_not_installed_on_container(container, system_pkg)


@pytest.mark.integration
@pytest.mark.parametrize(
    "binary, python_version",
    [
        (binary, python_version)
        for binary in [True, False]
        for python_version in SUPPORTED_PYTHON_VERSIONS
    ],
)
def test_control_truss_patch_ignored_changes(
    binary, python_version, control_model_handle_tag_tuple
):
    custom_model_control, th, tag = control_model_handle_tag_tuple
    th.update_python_version(python_version)

    def predict_with_ignored_changes():
        top_pycache_path = custom_model_control / "__pycache__"
        top_pycache_path.mkdir()
        (top_pycache_path / "bla.pyc").touch()
        model_pycache_path = custom_model_control / "model" / "__pycache__"
        model_pycache_path.mkdir()
        (model_pycache_path / "foo.pyc").touch()
        return th.docker_predict([1], tag=tag, binary=binary)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag, binary=binary)
        orig_num_truss_images = current_num_docker_images(th)

        predict_with_ignored_changes()
        assert current_num_docker_images(th) == orig_num_truss_images


# todo(abu/justin) remove once ignored
@pytest.mark.skip(reason="Unsupported patch")
@pytest.mark.integration
def test_patch_added_model_dir(
    custom_model_control, tmp_path, control_model_handle_tag_tuple
):
    custom_model_control, th, tag = control_model_handle_tag_tuple

    def predict_with_added_model_dir_file():
        code_file_dir = custom_model_control / "model" / "dir"
        code_file_dir.mkdir(parents=True)
        with (code_file_dir / "foo.bar").open("w") as model_code_file:
            model_code_file.write("foobar")
        return th.docker_predict([1], build_dir=tmp_path, tag=tag)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag)
        orig_num_truss_images = len(th.get_all_docker_images())

        predict_with_added_model_dir_file()
        assert orig_num_truss_images == current_num_docker_images(th)
        assert (tmp_path / "model" / "dir" / "foo.bar").exists()


# todo(abu/justin) remove once ignored
@pytest.mark.skip(reason="Unsupported patch")
@pytest.mark.integration
def test_patch_data_dir(control_model_handle_tag_tuple):
    custom_model_control, th, tag = control_model_handle_tag_tuple

    def predict_with_data_dir_change():
        path = custom_model_control / "data" / "dummy"
        path.touch()
        th.docker_predict([1], tag=tag)
        with path.open("w") as file:
            file.write("foobar")
        th.docker_predict([1], tag=tag)
        path.unlink()
        return th.docker_predict([1], tag=tag)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag)
        orig_num_truss_images = len(th.get_all_docker_images())

        predict_with_data_dir_change()
        assert orig_num_truss_images == current_num_docker_images(th)


# todo(abu/justin) remove once ignored
@pytest.mark.skip(reason="Unsupported patch")
@pytest.mark.integration
def test_patch_env_var(control_model_handle_tag_tuple):
    _, th, tag = control_model_handle_tag_tuple

    def predict_with_environment_variables_change():
        th.add_environment_variable("foo", "bar")
        th.docker_predict([1], tag=tag)
        verify_environment_variable_on_container(
            th.get_running_serving_container_ignore_hash(), "foo", "bar"
        )
        th.add_environment_variable("foo", "bar2")
        th.kill_container()
        th.docker_predict([1], tag=tag)
        verify_environment_variable_on_container(
            th.get_running_serving_container_ignore_hash(), "foo", "bar2"
        )
        th.clear_environment_variables()
        th.kill_container()
        th.docker_predict([1], tag=tag)
        with pytest.raises(AssertionError):
            verify_environment_variable_on_container(
                th.get_running_serving_container_ignore_hash(), "foo", "bar2"
            )

    with ensure_kill_all():
        th.docker_predict([1], tag=tag)
        orig_num_truss_images = len(th.get_all_docker_images())

        predict_with_environment_variables_change()
        assert orig_num_truss_images == current_num_docker_images(th)


# todo(abu/justin) remove once ignored
@pytest.mark.skip(reason="Unsupported patch")
@pytest.mark.integration
def test_patch_external_package_dirs(custom_model_with_external_package):
    th = TrussHandle(custom_model_with_external_package)
    tag = "test-docker-custom-model-control-external-package-tag:0.0.1"

    with ensure_kill_all():
        th.clear_external_packages()
        new_model_code = """
class Model:
    def predict(self, model_input):
        return [2 for i in model_input]
        """
        model_code_file_path = custom_model_with_external_package / "model" / "model.py"
        with model_code_file_path.open("w") as model_code_file:
            model_code_file.write(new_model_code)
        th.docker_predict([1], tag=tag)
        th.kill_container()
        orig_num_truss_images = len(th.get_all_docker_images())
        th.add_external_package("../ext_pkg")
        th.add_external_package("../ext_pkg2")
        with model_code_file_path.open("w") as model_code_file:
            model_code_file.write(CUSTOM_MODEL_USING_EXTERNAL_PACKAGE_CODE)
        th.docker_predict([1], tag=tag)
        assert orig_num_truss_images == current_num_docker_images(th)
        assert (custom_model_with_external_package / "ext_pkg").exists() and (
            custom_model_with_external_package / "ext_pkg2"
        ).exists()


# todo(abu/justin) remove once ignored
@pytest.mark.skip(reason="Unsupported patch")
@pytest.mark.integration
def test_patch_secrets(control_model_handle_tag_tuple):
    _, th, tag = control_model_handle_tag_tuple

    def predict_with_secrets():
        th.add_secret("foo", "bar")
        return th.docker_predict([1], tag=tag)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag)
        orig_num_truss_images = len(th.get_all_docker_images())

        predict_with_secrets()
        assert orig_num_truss_images == current_num_docker_images(th)


# todo(abu/justin) remove once ignored
@pytest.mark.skip(reason="Unsupported patch")
@pytest.mark.integration
def test_predict_with_external_data_change(
    custom_model_external_data_access_tuple_fixture, tmp_path
):
    truss_dir, _ = custom_model_external_data_access_tuple_fixture
    th = TrussHandle(truss_dir)
    tag = "test-external-data-access-tag:0.0.1"
    with ensure_kill_all():
        th.docker_predict([], tag=tag)
        orig_num_truss_images = len(th.get_all_docker_images())
        th.remove_all_external_data()
        new_model_code = """
class Model:
    def __init__(self, data_dir):
        self._data_dir = data_dir
        pass

    def predict(self, model_input):
        return None
"""
        model_code_file_path = truss_dir / "model" / "model.py"
        with model_code_file_path.open("w") as model_code_file:
            model_code_file.write(new_model_code)
        th.docker_predict([], tag=tag)
        content = "foobar"
        filename = "foobar.txt"
        (tmp_path / filename).write_text(content)
        port = 9090
        proc = subprocess.Popen(
            ["python", "-m", "http.server", str(port), "--bind", "*"],
            cwd=tmp_path,
        )
        new_model_code = """
class Model:
    def __init__(self, data_dir):
        self._data_dir = data_dir
        pass

    def predict(self, model_input):
        with (self._data_dir / 'foobar.txt').open() as file:
            return file.read()
"""
        model_code_file_path = truss_dir / "model" / "model.py"
        with model_code_file_path.open("w") as model_code_file:
            model_code_file.write(new_model_code)
        try:
            url = f"http://localhost:{port}/{filename}"
            th.add_external_data_item(url, filename)
            result = th.docker_predict([], tag=tag)
            assert (
                result == content
                and orig_num_truss_images == current_num_docker_images(th)
            )
        finally:
            proc.kill()



import logging
import tempfile
import time
from pathlib import Path
from threading import Thread

import numpy as np
import pytest
import requests
from requests.exceptions import RequestException
from truss.constants import PYTORCH
from truss.model_frameworks import SKLearn
from truss.model_inference import (
    infer_model_information,
    map_to_supported_python_version,
    validate_provided_parameters_with_model,
)
from truss.tests.test_testing_utilities_for_other_tests import ensure_kill_all
from truss.truss_handle import TrussHandle

logger = logging.getLogger(__name__)


def test_pytorch_init_arg_validation(
    pytorch_model_with_init_args, pytorch_model_init_args
):
    pytorch_model_with_init_args, _ = pytorch_model_with_init_args
    # Validates with args and kwargs
    validate_provided_parameters_with_model(
        pytorch_model_with_init_args.__class__, pytorch_model_init_args
    )

    # Errors if bad args
    with pytest.raises(ValueError):
        validate_provided_parameters_with_model(
            pytorch_model_with_init_args.__class__, {"foo": "bar"}
        )

    # Validates with only args
    copied_args = pytorch_model_init_args.copy()
    copied_args.pop("kwarg1")
    copied_args.pop("kwarg2")
    validate_provided_parameters_with_model(pytorch_model_with_init_args, copied_args)

    # Requires all args
    with pytest.raises(ValueError):
        validate_provided_parameters_with_model(pytorch_model_with_init_args, {})


def test_infer_model_information(pytorch_model_with_init_args):
    model_info = infer_model_information(pytorch_model_with_init_args[0])
    assert model_info.model_framework == PYTORCH
    assert model_info.model_type == "MyModel"


@pytest.mark.parametrize(
    "python_version, expected_python_version",
    [
        ("py37", "py38"),
        ("py38", "py38"),
        ("py39", "py39"),
        ("py310", "py310"),
        ("py311", "py311"),
        ("py312", "py311"),
        ("py36", "py38"),
    ],
)
def test_map_to_supported_python_version(python_version, expected_python_version):
    out_python_version = map_to_supported_python_version(python_version)
    assert out_python_version == expected_python_version


@pytest.mark.integration
def test_binary_request(sklearn_rfc_model):
    with ensure_kill_all(), tempfile.TemporaryDirectory(dir=".") as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, "truss")
        sklearn_framework = SKLearn()
        sklearn_framework.to_truss(sklearn_rfc_model, truss_dir)
        tr = TrussHandle(truss_dir)
        predictions = tr.docker_predict([[0, 0, 0, 0]], local_port=8090, binary=True)
        assert len(predictions["probabilities"]) == 1
        assert np.shape(predictions["probabilities"]) == (1, 3)


@pytest.mark.integration
def test_model_load_failure_truss():
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / "model_load_failure_test"
        tr = TrussHandle(truss_dir)

        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=False)

        # Sleep a few seconds to get the server some time to  wake up
        time.sleep(10)

        truss_server_addr = "http://localhost:8090"

        def handle_request_exception(func):
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except RequestException:
                    return False

            return wrapper

        @handle_request_exception
        def _test_liveness_probe(expected_code):
            live = requests.get(f"{truss_server_addr}/")
            assert live.status_code == expected_code
            return True

        @handle_request_exception
        def _test_readiness_probe(expected_code):
            ready = requests.get(f"{truss_server_addr}/v1/models/model")
            assert ready.status_code == expected_code
            return True

        @handle_request_exception
        def _test_ping(expected_code):
            ping = requests.get(f"{truss_server_addr}/ping")
            assert ping.status_code == expected_code
            return True

        @handle_request_exception
        def _test_invocations(expected_code):
            invocations = requests.post(f"{truss_server_addr}/invocations", json={})
            assert invocations.status_code == expected_code
            return True

        # The server should be completely down so all requests should result in a RequestException.
        # The decorator handle_request_exception catches the RequestException and returns False.
        assert not _test_readiness_probe(expected_code=200)
        assert not _test_liveness_probe(expected_code=200)
        assert not _test_ping(expected_code=200)
        assert not _test_invocations(expected_code=200)


@pytest.mark.integration
def test_slow_truss():
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / "server_conformance_test_truss"
        tr = TrussHandle(truss_dir)

        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=False)

        truss_server_addr = "http://localhost:8090"

        def _test_liveness_probe(expected_code):
            live = requests.get(f"{truss_server_addr}/")
            assert live.status_code == expected_code

        def _test_readiness_probe(expected_code):
            ready = requests.get(f"{truss_server_addr}/v1/models/model")
            assert ready.status_code == expected_code

        def _test_ping(expected_code):
            ping = requests.get(f"{truss_server_addr}/ping")
            assert ping.status_code == expected_code

        def _test_invocations(expected_code):
            invocations = requests.post(f"{truss_server_addr}/invocations", json={})
            assert invocations.status_code == expected_code

        SERVER_WARMUP_TIME = 3
        LOAD_TEST_TIME = 12
        LOAD_BUFFER_TIME = 7
        PREDICT_TEST_TIME = 15

        # Sleep a few seconds to get the server some time to wake up
        time.sleep(SERVER_WARMUP_TIME)

        # The truss takes about 30 seconds to load.
        # We want to make sure that it's not ready for that time.
        for _ in range(LOAD_TEST_TIME):
            _test_liveness_probe(200)
            _test_readiness_probe(503)
            _test_ping(503)
            _test_invocations(503)
            time.sleep(1)

        time.sleep(LOAD_BUFFER_TIME)
        _test_liveness_probe(200)
        _test_readiness_probe(200)
        _test_ping(200)

        predict_call = Thread(
            target=lambda: requests.post(
                f"{truss_server_addr}/v1/models/model:predict", json={}
            )
        )
        predict_call.start()

        for _ in range(PREDICT_TEST_TIME):
            _test_liveness_probe(200)
            _test_readiness_probe(200)
            _test_ping(200)
            time.sleep(1)

        predict_call.join()

        _test_invocations(200)
